using CallFlow.CFD;
using CallFlow;
using MimeKit;
using MySql.Data.MySqlClient;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json;
using Npgsql;
using System.Collections.Generic;
using System.Data.Common;
using System.Data.SqlClient;
using System.Data;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks.Dataflow;
using System.Threading.Tasks;
using System.Threading;
using System;
using TCX.Configuration;

namespace LoopCalls
{
    public class Main : ScriptBase<Main>, ICallflow, ICallflowProcessor
    {
        private bool executionStarted;
        private bool executionFinished;

        private BufferBlock<AbsEvent> eventBuffer;

        private int currentComponentIndex;
        private List<AbsComponent> mainFlowComponentList;
        private List<AbsComponent> disconnectFlowComponentList;
        private List<AbsComponent> errorFlowComponentList;
        private List<AbsComponent> currentFlowComponentList;

        private LogFormatter logFormatter;
        private TimerManager timerManager;
        private Dictionary<string, Variable> variableMap;
        private TempWavFileManager tempWavFileManager;
        private PromptQueue promptQueue;

        private void DisconnectCallAndExitCallflow()
        {
            logFormatter.Trace("Callflow finished, disconnecting call...");
            MyCall.Terminate();
        }

        private async Task ExecuteErrorFlow()
        {
            if (currentFlowComponentList == errorFlowComponentList)
            {
                logFormatter.Trace("Error during error handler flow, exiting callflow...");
                DisconnectCallAndExitCallflow();
            }
            else
            {
                currentFlowComponentList = errorFlowComponentList;
                currentComponentIndex = 0;
                if (errorFlowComponentList.Count > 0)
                {
                    logFormatter.Trace("Start executing error handler flow...");
                    await ProcessStart();
                }
                else
                {
                    logFormatter.Trace("Error handler flow is empty...");
                    DisconnectCallAndExitCallflow();
                }
            }
        }

        private EventResults CheckEventResult(EventResults eventResult)
        {
            if (eventResult == EventResults.MoveToNextComponent && ++currentComponentIndex == currentFlowComponentList.Count)
            {
                DisconnectCallAndExitCallflow();
                return EventResults.Exit;
            }
            else if (eventResult == EventResults.Exit)
                DisconnectCallAndExitCallflow();

            return eventResult;
        }

        private void InitializeVariables(string callID)
        {
            // Call variables
            variableMap["session.ani"] = new Variable(MyCall.Caller.CallerID);
            variableMap["session.callid"] = new Variable(callID);
            variableMap["session.dnis"] = new Variable(MyCall.DN.Number);
            variableMap["session.did"] = new Variable(MyCall.Caller.CalledNumber);
            variableMap["session.audioFolder"] = new Variable(Path.Combine(RecordingManager.Instance.AudioFolder, promptQueue.ProjectAudioFolder));
            variableMap["session.transferingExtension"] = new Variable(MyCall["onbehlfof"] ?? string.Empty);

            // Standard variables
            variableMap["RecordResult.NothingRecorded"] = new Variable(RecordComponent.RecordResults.NothingRecorded);
            variableMap["RecordResult.StopDigit"] = new Variable(RecordComponent.RecordResults.StopDigit);
            variableMap["RecordResult.Completed"] = new Variable(RecordComponent.RecordResults.Completed);
            variableMap["MenuResult.Timeout"] = new Variable(MenuComponent.MenuResults.Timeout);
            variableMap["MenuResult.InvalidOption"] = new Variable(MenuComponent.MenuResults.InvalidOption);
            variableMap["MenuResult.ValidOption"] = new Variable(MenuComponent.MenuResults.ValidOption);
            variableMap["UserInputResult.Timeout"] = new Variable(UserInputComponent.UserInputResults.Timeout);
            variableMap["UserInputResult.InvalidDigits"] = new Variable(UserInputComponent.UserInputResults.InvalidDigits);
            variableMap["UserInputResult.ValidDigits"] = new Variable(UserInputComponent.UserInputResults.ValidDigits);

            // User variables
            variableMap["callflow$.PIN"] = new Variable("");
            variableMap["callflow$.Ext_No"] = new Variable("");
            variableMap["callflow$.Destination"] = new Variable("");
            variableMap["callflow$.phoneNum"] = new Variable("");
            variableMap["callflow$.DNIS"] = new Variable("");
            variableMap["callflow$.transferext"] = new Variable("");
            variableMap["callflow$.callid"] = new Variable("");
            variableMap["callflow$.ReturnValue"] = new Variable("");
            variableMap["RecordResult.NothingRecorded"] = new Variable(RecordComponent.RecordResults.NothingRecorded);
            variableMap["RecordResult.StopDigit"] = new Variable(RecordComponent.RecordResults.StopDigit);
            variableMap["RecordResult.Completed"] = new Variable(RecordComponent.RecordResults.Completed);
            variableMap["MenuResult.Timeout"] = new Variable(MenuComponent.MenuResults.Timeout);
            variableMap["MenuResult.InvalidOption"] = new Variable(MenuComponent.MenuResults.InvalidOption);
            variableMap["MenuResult.ValidOption"] = new Variable(MenuComponent.MenuResults.ValidOption);
            variableMap["UserInputResult.Timeout"] = new Variable(UserInputComponent.UserInputResults.Timeout);
            variableMap["UserInputResult.InvalidDigits"] = new Variable(UserInputComponent.UserInputResults.InvalidDigits);
            variableMap["UserInputResult.ValidDigits"] = new Variable(UserInputComponent.UserInputResults.ValidDigits);
            
        }

        private void InitializeComponents(ICallflow callflow, ICall myCall, string logHeader)
        {
            {
            PromptPlaybackComponent PromptPlayback1 = new PromptPlaybackComponent("PromptPlayback1", callflow, myCall, logHeader);
            PromptPlayback1.AllowDtmfInput = true;
            PromptPlayback1.Prompts.Add(new AudioFilePrompt(() => { return "beep.wav"; }));
            mainFlowComponentList.Add(PromptPlayback1);
            TransferComponent CallConnect = new TransferComponent("CallConnect", callflow, myCall, logHeader);
            CallConnect.DestinationHandler = () => { return Convert.ToString(variableMap["session.ani"].Value); };
            CallConnect.DelayMilliseconds = 500;
            mainFlowComponentList.Add(CallConnect);
            DisconnectCallComponent DisconnectCall2 = new DisconnectCallComponent("DisconnectCall2", callflow, myCall, logHeader);
            mainFlowComponentList.Add(DisconnectCall2);
            }
            {
            }
            {
            }
            

            // Add a final DisconnectCall component to the main and error handler flows, in order to complete pending prompt playbacks...
            DisconnectCallComponent mainAutoAddedFinalDisconnectCall = new DisconnectCallComponent("mainAutoAddedFinalDisconnectCall", callflow, myCall, logHeader);
            DisconnectCallComponent errorHandlerAutoAddedFinalDisconnectCall = new DisconnectCallComponent("errorHandlerAutoAddedFinalDisconnectCall", callflow, myCall, logHeader);
            mainFlowComponentList.Add(mainAutoAddedFinalDisconnectCall);
            errorFlowComponentList.Add(errorHandlerAutoAddedFinalDisconnectCall);
        }

        private bool IsServerOfficeHourActive(ICall myCall)
        {
            DateTime nowDt = DateTime.Now;
            Tenant tenant = myCall.PS.GetTenant();
            if (tenant == null || tenant.IsHoliday(new DateTimeOffset(nowDt)))
                return false;

            string overrideOfficeTime = tenant.GetPropertyValue("OVERRIDEOFFICETIME");
            if (!String.IsNullOrEmpty(overrideOfficeTime))
            {
                if (overrideOfficeTime == "1") // Forced to in office hours
                    return true;
                else if (overrideOfficeTime == "2") // Forced to out of office hours
                    return false;
            }

            Schedule officeHours = tenant.Hours;
            Nullable<bool> result = officeHours.IsActiveTime(nowDt);
            return result.GetValueOrDefault(false);
        }

        public Main()
        {
            this.executionStarted = false;
            this.executionFinished = false;

            this.eventBuffer = new BufferBlock<AbsEvent>();

            this.currentComponentIndex = 0;
            this.mainFlowComponentList = new List<AbsComponent>();
            this.disconnectFlowComponentList = new List<AbsComponent>();
            this.errorFlowComponentList = new List<AbsComponent>();
            this.currentFlowComponentList = mainFlowComponentList;

            this.timerManager = new TimerManager();
            this.timerManager.OnTimeout += (state) => eventBuffer.Post(new TimeoutEvent(state));
            this.variableMap = new Dictionary<string, Variable>();
        }

        public override void Start()
        {
            MyCall.SetBackgroundAudio(false, new string[] { });

            string callID = MyCall?.Caller["chid"] ?? "Unknown";
            string logHeader = $"LoopCalls - CallID {callID}";
            this.logFormatter = new LogFormatter(MyCall, logHeader, "Callflow");
            this.promptQueue = new PromptQueue(this, MyCall, "LoopCalls", logHeader);
            this.tempWavFileManager = new TempWavFileManager(logFormatter);
            this.timerManager.CallStarted();

            InitializeComponents(this, MyCall, logHeader);
            InitializeVariables(callID);
            
            MyCall.OnTerminated += () => eventBuffer.Post(new CallTerminatedEvent());
            MyCall.OnDTMFInput += x => eventBuffer.Post(new DTMFReceivedEvent(x));

            logFormatter.Trace("Start executing main flow...");
            eventBuffer.Post(new StartEvent());
            Task.Run(() => EventProcessingLoop());

            
        }
        
        public void PostStartEvent()
        {
            eventBuffer.Post(new StartEvent());
        }

        public void PostDTMFReceivedEvent(char digit)
        {
            eventBuffer.Post(new DTMFReceivedEvent(digit));
        }

        public void PostPromptPlayedEvent()
        {
            eventBuffer.Post(new PromptPlayedEvent());
        }

        public void PostTransferFailedEvent()
        {
            eventBuffer.Post(new TransferFailedEvent());
        }

        public void PostMakeCallResultEvent(bool result)
        {
            eventBuffer.Post(new MakeCallResultEvent(result));
        }

        public void PostCallTerminatedEvent()
        {
            eventBuffer.Post(new CallTerminatedEvent());
        }

        public void PostTimeoutEvent(object state)
        {
            eventBuffer.Post(new TimeoutEvent(state));
        }

        private async Task EventProcessingLoop()
        {
            executionStarted = true;
            while (!executionFinished)
            {
                AbsEvent evt = await eventBuffer.ReceiveAsync();
                await evt?.ProcessEvent(this);
            }
        }
        
        public async Task ProcessStart()
        {
            try
            {
                EventResults eventResult;
                do
                {
                    AbsComponent currentComponent = currentFlowComponentList[currentComponentIndex];
                    logFormatter.Trace("Start executing component '" + currentComponent.Name + "'");
                    eventResult = await currentComponent.Start(timerManager, variableMap, tempWavFileManager, promptQueue);
                }
                while (CheckEventResult(eventResult) == EventResults.MoveToNextComponent);
            }
            catch (Exception exc)
            {
                logFormatter.Error("Error executing last component: " + exc.ToString());
                await ExecuteErrorFlow();
            }
        }

        public async Task ProcessDTMFReceived(char digit)
        {
            try
            {
                AbsComponent currentComponent = currentFlowComponentList[currentComponentIndex];
                logFormatter.Trace("OnDTMFReceived for component '" + currentComponent.Name + "' - Digit: '" + digit + "'");
                EventResults eventResult = await currentComponent.OnDTMFReceived(timerManager, variableMap, tempWavFileManager, promptQueue, digit);
                if (CheckEventResult(eventResult) == EventResults.MoveToNextComponent)
                    await ProcessStart();
            }
            catch (Exception exc)
            {
                logFormatter.Error("Error executing last component: " + exc.ToString());
                await ExecuteErrorFlow();
            }
        }

        public async Task ProcessPromptPlayed()
        {
            try
            {
                promptQueue.NotifyPlayFinished();
                AbsComponent currentComponent = currentFlowComponentList[currentComponentIndex];
                logFormatter.Trace("OnPromptPlayed for component '" + currentComponent.Name + "'");
                EventResults eventResult = await currentComponent.OnPromptPlayed(timerManager, variableMap, tempWavFileManager, promptQueue);
                if (CheckEventResult(eventResult) == EventResults.MoveToNextComponent)
                    await ProcessStart();
            }
            catch (Exception exc)
            {
                logFormatter.Error("Error executing last component: " + exc.ToString());
                await ExecuteErrorFlow();
            }
        }

        public async Task ProcessTransferFailed()
        {
            try
            {
                AbsComponent currentComponent = currentFlowComponentList[currentComponentIndex];
                logFormatter.Trace("OnTransferFailed for component '" + currentComponent.Name + "'");
                EventResults eventResult = await currentComponent.OnTransferFailed(timerManager, variableMap, tempWavFileManager, promptQueue);
                if (CheckEventResult(eventResult) == EventResults.MoveToNextComponent)
                    await ProcessStart();
            }
            catch (Exception exc)
            {
                logFormatter.Error("Error executing last component: " + exc.ToString());
                await ExecuteErrorFlow();
            }
        }

        public async Task ProcessMakeCallResult(bool result)
        {
            try
            {
                AbsComponent currentComponent = currentFlowComponentList[currentComponentIndex];
                logFormatter.Trace("OnMakeCallResult for component '" + currentComponent.Name + "' - Result: '" + result + "'");
                EventResults eventResult = await currentComponent.OnMakeCallResult(timerManager, variableMap, tempWavFileManager, promptQueue, result);
                if (CheckEventResult(eventResult) == EventResults.MoveToNextComponent)
                    await ProcessStart();
            }
            catch (Exception exc)
            {
                logFormatter.Error("Error executing last component: " + exc.ToString());
                await ExecuteErrorFlow();
            }
        }

        public async Task ProcessCallTerminated()
        {
            try
            {
                if (executionStarted)
                {
                    // First notify the current component
                    AbsComponent currentComponent = currentFlowComponentList[currentComponentIndex];
                    logFormatter.Trace("OnCallTerminated for component '" + currentComponent.Name + "'");
                    await currentComponent.OnCallTerminated(timerManager, variableMap, tempWavFileManager, promptQueue);

                    // Next, execute disconnect flow
                    currentFlowComponentList = disconnectFlowComponentList;
                    logFormatter.Trace("Start executing disconnect handler flow...");
                    foreach (AbsComponent component in disconnectFlowComponentList)
                    {
                        logFormatter.Trace("Start executing component '" + component.Name + "'");
                        await component.Start(timerManager, variableMap, tempWavFileManager, promptQueue);
                    }
                }
            }
            catch (Exception exc)
            {
                logFormatter.Error("Error executing last component: " + exc.ToString());
                await ExecuteErrorFlow();
            }
            finally
            {
                // Finally, delete temporary files
                tempWavFileManager.DeleteFilesAndFolders();
                executionFinished = true;
            }
        }

        public async Task ProcessTimeout(object state)
        {
            try
            {
                AbsComponent currentComponent = currentFlowComponentList[currentComponentIndex];
                logFormatter.Trace("OnTimeout for component '" + currentComponent.Name + "'");
                EventResults eventResult = await currentComponent.OnTimeout(timerManager, variableMap, tempWavFileManager, promptQueue, state);
                if (CheckEventResult(eventResult) == EventResults.MoveToNextComponent)
                    await ProcessStart();
            }
            catch (Exception exc)
            {
                logFormatter.Error("Error executing last component: " + exc.ToString());
                await ExecuteErrorFlow();
            }
        }


        public class TcxSetExtensionStatusTempComponent : AbsComponent
        {
            private StringExpressionHandler extensionHandler = null;
            private StringExpressionHandler profileNameHandler = null;

            public TcxSetExtensionStatusTempComponent(string name, ICallflow callflow, ICall myCall, string logHeader) : base(name, callflow, myCall, logHeader)
            {
            }

            public string Extension { get; set; } = string.Empty;
            public string ProfileName { get; set; } = string.Empty;

            public StringExpressionHandler ExtensionHandler
            {
                set { extensionHandler = value; }
            }

            public StringExpressionHandler ProfileNameHandler
            {
                set { profileNameHandler = value; }
            }

            private EventResults ExecuteStart()
            {
                if (extensionHandler != null) Extension = extensionHandler();
                if (profileNameHandler != null) ProfileName = profileNameHandler();

                using (DN dn = myCall.PS.GetDNByNumber(Extension))
                {
                    if (!(dn is Extension ext)) throw new Exception("Extension '" + Extension + "' could not be found.");

                    logFormatter.Trace("Setting status for Extension='" + Extension + "' to '" + ProfileName + "'");

                    // Activate the selected profile
                    bool profileFound = false;
                    foreach (FwdProfile profile in ext.FwdProfiles)
                    {
                        if (profile.Name == ProfileName)
                        {
                            ext.CurrentProfile = profile;
                            profileFound = true;
                            break;
                        }
                    }

                    if (profileFound)
                    {
                        // And disable override if active
                        if (ext.IsOverrideActiveNow) ext.OverrideExpiresAt = DateTime.UtcNow;
                        logFormatter.Trace("Status for Extension='" + Extension + "' updated.");
                        ext.Save();
                    }
                    else
                        logFormatter.Trace("Status for Extension='" + Extension + "' could not be updated, profile name not found.");
                }
                return EventResults.MoveToNextComponent;
            }

            public override Task<EventResults> Start(Dictionary<string, Variable> variableMap)
            {
                return Task.FromResult(ExecuteStart());
            }

            public override Task<EventResults> Start(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return Task.FromResult(ExecuteStart());
            }

            public override Task<EventResults> OnDTMFReceived(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue, char digit)
            {
                return IgnoreDTMFReceivedEventWithResult(digit, EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnPromptPlayed(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return IgnoreEventWithResult("OnPromptPlayed", EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnTransferFailed(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return IgnoreEventWithResult("OnTransferFailed", EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnMakeCallResult(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue, bool result)
            {
                return IgnoreMakeCallResult(result, EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnCallTerminated(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return ProcessEventWithResult("OnCallTerminated", EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnTimeout(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue, object state)
            {
                return IgnoreEventWithResult("OnTimeout", EventResults.MoveToNextComponent);
            }
        }

        public class MySqlDatabaseAccessTempComponent : DatabaseAccessTempComponent
        {
            protected override DbConnection CreateConnection()
            {
                return new MySqlConnection(useConnectionString ? connectionString : string.Format("Server={0};Port={1};Database={2};User Id={3};Password={4}", server, port == -1 ? 3306 : port, database, userName, password));
            }

            public MySqlDatabaseAccessTempComponent(string name, ICallflow callflow, ICall myCall, string logHeader) : base(name, callflow, myCall, logHeader)
            {
            }
        }

        public class PostgresqlDatabaseAccessTempComponent : DatabaseAccessTempComponent
        {
            protected override DbConnection CreateConnection()
            {
                return new NpgsqlConnection(useConnectionString ? connectionString : string.Format("Server={0};Port={1};Database={2};User Id={3};Password={4}", server, port == -1 ? 5432 : port, database, userName, password));
            }

            public PostgresqlDatabaseAccessTempComponent(string name, ICallflow callflow, ICall myCall, string logHeader) : base(name, callflow, myCall, logHeader)
            {
            }
        }

        public class SqlServerDatabaseAccessTempComponent : DatabaseAccessTempComponent
        {
            protected override DbConnection CreateConnection()
            {
                return new SqlConnection(useConnectionString ? connectionString : string.Format("Data Source=tcp:{0}{1};initial catalog={2};user id={3};password={4};Pooling=true", server, port == -1 ? string.Empty : "," + port, database, userName, password));
            }

            public SqlServerDatabaseAccessTempComponent(string name, ICallflow callflow, ICall myCall, string logHeader) : base(name, callflow, myCall, logHeader)
            {
            }
        }

        public abstract class DatabaseAccessTempComponent : AbsComponent
        {
            public enum StatementTypes
            {
                Query,
                NonQuery,
                Scalar
            }

            protected bool useConnectionString = false;
            protected string connectionString = string.Empty;
            protected string server = string.Empty;
            protected int port = -1;
            protected string database = string.Empty;
            protected string userName = string.Empty;
            protected string password = string.Empty;
            private string sqlStatement = string.Empty;

            private StringExpressionHandler connectionStringHandler = null;
            private StringExpressionHandler serverHandler = null;
            private IntegerExpressionHandler portHandler = null;
            private StringExpressionHandler databaseHandler = null;
            private StringExpressionHandler userNameHandler = null;
            private StringExpressionHandler passwordHandler = null;
            private StringExpressionHandler sqlStatementHandler = null;

            protected abstract DbConnection CreateConnection();

            private async Task ExecuteStatement(DbConnection conn)
            {
                if (StatementType == StatementTypes.Query)
                {
                    using (DbDataReader reader = await ExecuteReaderAsync(conn))
                    {
                        while (await reader.ReadAsync())
                        {
                            object[] recordValues = new object[reader.FieldCount];
                            reader.GetValues(recordValues);
                            QueryResult.Add(recordValues);
                        }
                    }
                }
                else if (StatementType == StatementTypes.NonQuery)
                    NonQueryResult = await ExecuteNonQueryAsync(conn);
                else
                    ScalarResult = await ExecuteScalarAsync(conn);
            }

            private async Task<DbCommand> PrepareCommandAsync(DbConnection conn)
            {
                logFormatter.Trace("Command to execute: " + sqlStatement);

                await conn.OpenAsync();
                DbCommand cmd = conn.CreateCommand();
                cmd.CommandText = sqlStatement;
                cmd.CommandTimeout = Timeout;
                cmd.CommandType = CommandType.Text;
                return cmd;
            }

            private async Task<DbDataReader> ExecuteReaderAsync(DbConnection conn)
            {
                DbCommand cmd = await PrepareCommandAsync(conn);
                return await cmd.ExecuteReaderAsync();
            }

            private async Task<int> ExecuteNonQueryAsync(DbConnection conn)
            {
                DbCommand cmd = await PrepareCommandAsync(conn);
                return await cmd.ExecuteNonQueryAsync();
            }

            private async Task<object> ExecuteScalarAsync(DbConnection conn)
            {
                DbCommand cmd = await PrepareCommandAsync(conn);
                return await cmd.ExecuteScalarAsync();
            }

            public DatabaseAccessTempComponent(string name, ICallflow callflow, ICall myCall, string logHeader) : base(name, callflow, myCall, logHeader)
            {
            }

            public bool UseConnectionString
            {
                get { return useConnectionString; }
                set { useConnectionString = value; }
            }

            public string ConnectionString
            {
                get { return connectionString; }
                set { connectionString = value; }
            }

            public string Server
            {
                get { return server; }
                set { server = value; }
            }

            public int Port
            {
                get { return port; }
                set { port = value; }
            }

            public string Database
            {
                get { return database; }
                set { database = value; }
            }

            public string UserName
            {
                get { return userName; }
                set { userName = value; }
            }

            public string Password
            {
                get { return password; }
                set { password = value; }
            }

            public string SqlStatement
            {
                get { return sqlStatement; }
                set { sqlStatement = value; }
            }

            public StatementTypes StatementType { get; set; } = StatementTypes.Query;
            public int Timeout { get; set; } = 30000;

            public StringExpressionHandler ConnectionStringHandler
            {
                set { connectionStringHandler = value; }
            }

            public StringExpressionHandler ServerHandler
            {
                set { serverHandler = value; }
            }

            public IntegerExpressionHandler PortHandler
            {
                set { portHandler = value; }
            }

            public StringExpressionHandler DatabaseHandler
            {
                set { databaseHandler = value; }
            }

            public StringExpressionHandler UserNameHandler
            {
                set { userNameHandler = value; }
            }

            public StringExpressionHandler PasswordHandler
            {
                set { passwordHandler = value; }
            }

            public StringExpressionHandler SqlStatementHandler
            {
                set { sqlStatementHandler = value; }
            }

            public object ScalarResult { get; private set; } = null;
            public int NonQueryResult { get; private set; } = 0;
            public List<object[]> QueryResult { get; private set; } = new List<object[]>();

            private async Task<EventResults> ExecuteStart()
            {
                ScalarResult = null;
                NonQueryResult = 0;
                QueryResult = new List<object[]>();

                if (useConnectionString)
                {
                    if (connectionStringHandler != null) connectionString = connectionStringHandler();
                    logFormatter.Trace("Start executing component with connectionString='" + connectionString + "'");
                }
                else
                {
                    if (serverHandler != null) server = serverHandler();
                    if (portHandler != null) port = portHandler();
                    if (databaseHandler != null) database = databaseHandler();
                    if (userNameHandler != null) userName = userNameHandler();
                    if (passwordHandler != null) password = passwordHandler();

                    logFormatter.Trace("Start executing component with server='" + server + "' - port='" + port + "' - database='" + database + "' - statementType='" + StatementType + "'");
                }

                if (sqlStatementHandler != null) sqlStatement = sqlStatementHandler();

                using (DbConnection conn = CreateConnection())
                {
                    await ExecuteStatement(conn);
                }

                if (StatementType == StatementTypes.Query)
                {
                    StringBuilder sb = new StringBuilder();
                    foreach (object[] record in QueryResult)
                    {
                        sb.Append("[");
                        sb.Append(string.Join<object>(",", record));
                        sb.Append("]").AppendLine();
                    }
                    logFormatter.Trace("End executing component with queryResult: " + sb.ToString());
                }
                else if (StatementType == StatementTypes.Scalar)
                    logFormatter.Trace("End executing component with scalarResult: " + ScalarResult);
                else
                    logFormatter.Trace("End executing component with nonQueryResult: " + NonQueryResult);

                return EventResults.MoveToNextComponent;
            }

            public override async Task<EventResults> Start(Dictionary<string, Variable> variableMap)
            {
                return await ExecuteStart();
            }

            public override async Task<EventResults> Start(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return await ExecuteStart();
            }

            public override Task<EventResults> OnDTMFReceived(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue, char digit)
            {
                return IgnoreDTMFReceivedEventWithResult(digit, EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnPromptPlayed(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return IgnoreEventWithResult("OnPromptPlayed", EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnTransferFailed(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return IgnoreEventWithResult("OnTransferFailed", EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnMakeCallResult(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue, bool result)
            {
                return IgnoreMakeCallResult(result, EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnCallTerminated(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue)
            {
                return ProcessEventWithResult("OnCallTerminated", EventResults.MoveToNextComponent);
            }

            public override Task<EventResults> OnTimeout(TimerManager timerManager, Dictionary<string, Variable> variableMap, TempWavFileManager tempWavFileManager, PromptQueue promptQueue, object state)
            {
                return IgnoreEventWithResult("OnTimeout", EventResults.MoveToNextComponent);
            }
        }

        
class checkAvailableVMpin
{
    public int loopVMPin(string vmPin)
    {
        PhoneSystem ps = PhoneSystem.Root;
        int iAvailable = 0;
        foreach (Extension a in ps.GetTenants()[0].GetExtensions())
        {
            if (a.VMPIN != "")
            {
                if (a.VMEnabled == true)
                {
                    if (a.VMPIN == vmPin)
                    {
                        iAvailable = 1;
                        break;
                    }
                }
            }
        }
        return iAvailable;
    }
}


    class DESTINATION
    {
        public string GetDes(string strDestination)
        {
            string checkDestination = "";
            string _NumberDes = strDestination.Substring(4);
            return checkDestination;
        }
    }


    class PIN
    {
        public string Routing(string strExt, string strVMPIN)
        {
            string checkValidPIN = "";
            string _pinNumber = strVMPIN.Substring(0,4);

            try
            {
                PhoneSystem ps = PhoneSystem.Root;
                Extension extension = PhoneSystem.Root.GetDNByNumber(strExt) as Extension;
                if (extension.VMPIN == _pinNumber)
                {
                    checkValidPIN = "1";
                }
                else
                {
                    checkValidPIN = "0";
                }
            }
            catch (Exception ex)
            {

            }

            return checkValidPIN;
        }
    }


    }
}
